/*
 *  Command line utility for making sense of C declarations.
 *  Copyright © 2011 Nick Bowler
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <getopt.h>
#include "readline.h"
#include "cdecl.h"

static const char *progname = "cdecl99";
static const char sopts[] = "qbiVH";
static const struct option lopts[] = {
	{ "quiet",       0, NULL, 'q' },
	{ "batch",       0, NULL, 'b' },
	{ "interactive", 0, NULL, 'i' },
	{ "version",     0, NULL, 'V' },
	{ "help",        0, NULL, 'H' },
	{ 0 }
};

static void print_version(void)
{
	puts(PACKAGE_STRING);
	puts("Copyright (C) 2011 Nick Bowler.");
	puts("License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>.");
	puts("This is free software: you are free to change and redistribute it.");
	puts("There is NO WARRANTY, to the extent permitted by law.");
}

static void print_usage(FILE *f)
{
	fprintf(f, "Usage: %s [options]\n", progname);
}

static void print_help(void)
{
	print_usage(stdout);
	puts("Detailed help coming soon.");
}

static int cmd_explain(char *cmd, char *arg)
{
	static size_t bufsz;
	static char *buf;
	int ret = -1;

	struct cdecl *decl;
	size_t rc;

	decl = cdecl_parse_decl(arg);
	if (!decl)
		goto out;

	for (struct cdecl *i = decl; i; i = i->next) {
retry:
		rc = cdecl_explain(buf, bufsz, i);
		if (rc >= bufsz) {
			char *tmp;

			tmp = realloc(buf, rc + 1);
			if (!tmp) {
				fprintf(stderr, "failed to allocate memory\n");
				goto out;
			}

			buf = tmp;
			bufsz = rc + 1;
			goto retry;
		}

		printf("%s\n", buf);
	}

	ret = 1;
out:
	cdecl_free(decl);
	return ret;
}

static int cmd_quit(char *cmd, char *arg)
{
	return 0;
}

static int cmd_help(char *cmd, char *arg);

static const struct command {
	char name[16];
	int (*func)(char *cmd, char *arg);
	const char *blurb;
} commands[] = {
	{ "explain", cmd_explain, "Explain a C declaration." },
	{ "help",    cmd_help,    "Print this list of commands." },
	{ "quit",    cmd_quit,    "Quit the program." },
	{ "exit",    cmd_quit,    NULL }
};
static const size_t ncommands = sizeof commands / sizeof commands[0];

static int cmd_help(char *cmd, char *arg)
{
	for (size_t i = 0; i < ncommands; i++) {
		if (!commands[i].blurb)
			continue;

		printf("%s -- %s\n", commands[i].name, commands[i].blurb);
	}

	return 1;
}

static int repl(bool interactive)
{
	char *prompt = interactive ? "> " : NULL;
	int rc, ret = 0;
	char *line;

	for (; (line = readline(prompt)); free(line)) {
		char *cmd = line + strspn(line, " \t");
		char *arg = cmd  + strcspn(cmd, " \t");

		if (cmd[0] == '\0')
			continue;
		if (arg[0] != '\0')
			*arg++ = '\0';

		for (size_t i = 0; i < ncommands; i++) {
			if (strcmp(cmd, commands[i].name) != 0)
				continue;

			rc = commands[i].func(cmd, arg);
			if (!interactive && rc < 0)
				ret = -1;
			if (rc == 0)
				goto out;
			goto next;
		}

		fprintf(stderr, "Undefined command: %s\n", cmd);
		ret = -1;
	next:
		;
	}
out:
	free(line);
	return ret;
}

int main(int argc, char **argv)
{
	bool show_intro = true, interactive = true;
	int opt;

	if (argc > 0)
		progname = argv[0];

	while ((opt = getopt_long(argc, argv, sopts, lopts, NULL)) != -1) {
		switch (opt) {
		case 'q':
			show_intro = false;
			break;
		case 'b':
			interactive = false;
			break;
		case 'i':
			interactive = true;
			break;
		case 'V':
			print_version();
			return EXIT_SUCCESS;
		case 'H':
			print_help();
			return EXIT_SUCCESS;
		default:
			print_usage(stderr);
			return EXIT_FAILURE;
		}
	}

	if (interactive && show_intro)
		print_version();

	if (repl(interactive) != 0)
		return EXIT_FAILURE;

	return EXIT_SUCCESS;
}
